/* $Id: gTags.c,v 1.6 1999/04/11 17:00:37 ok Exp $ 
 *
 * gTags
 *
 * Author: Oliver Koell <mailto:ok@pobox.com>, <http://www.pobox.com/~ok>
 * 
 * USAGE:
 *
 * gTags requires ctags output, which it reads from STDIN - consult the 
 * README file for instructions. gTags accepts the following command line 
 * parameters:
 *
 *   gTags [-dlxv] [text]
 *
 *      -x  execute editor on tag selection (default).
 *      -l  print line number/regex to stdout on tag selection.
 *      -v  print version number to stdout.
 *      -d  print line number/regex and filename on tag selection (debug
 *          mode)
 *   the text argument will be displayed int gTags` main window title.
 */

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <string.h>

#include "gTags.h"


/* The default mode: execute the editor on tag selection (mirrors the 
 * command line options). Could also be 'l' or 'd'.
 */
#define DEFAULT_MODE 'x'

/* The invocation of the editor. This array will be passed to exec().
 * Before that, the "%key" and "%file" parameters are replaced with 
 * the appropriate values of the selected tag.
 * The default configuration will try to invoke NEdit in client mode, silently
 * starting a server, if none is running.
 */
static char  *exec_argv[] =  { "nc", 
                               "-noask", 
															 "-line", 
															 "%key",   /* replaced with line number/regex */
															 "%file",  /* replaced with file name */
															 NULL };

/* This is how Java tags are displayed.
 * The first element of a format identifies the type of a tag (as provided
 * by ctags' extension flags), the second is a sprintf() format string that
 * is applied to the identified tag.
 */
static tagFormat java_formats[] = { { 'f', "#%s" },   /* fields */
																		{ 'm', " %s()" },   /* methods */
	                                  { '\0', NULL }
                                  };

/* This is how c and c++ tags are displayed.
 */
static tagFormat c_formats[]    = { { 'd', "#%s" },   /* macro definitions */
																		{ 'v', " %s" },   /* variables */
																		{ 'f', "%s()" },  /* functions */
	                                  { '\0', NULL }
                                  };

/* maps the formats to filename extensions (languages).
 */
static langMap language_map[] = { { "java", java_formats },
                                  { "c",    c_formats },
                                  { "h",    c_formats },
                                  { NULL, NULL }
                                };


/* Optional window title from the command line text argument.
 */
static gchar *window_title = NULL;

/* execution mode (determined from command line).
 */
static char mode = DEFAULT_MODE;


/* Parses command line args
 */
static void parse_args( int argc, gchar *argv[] ) {
	char *option;
	
	while ( argc > 1 ) {
		argc--;
		if ( argv[argc][0] == '-' ) {
			option = argv[argc];
			while ( *(++option) ) {
				switch (*option) {
					case 'd': mode = *option;
					          break;
					case 'l': mode = *option;
					          break;
					case 'x': mode = *option;
					          break;
					case 'v': g_print("gTags version %s\n", GTAGS_VERSION);
					          exit( 0 );
					          break;
					default : g_print("Bad option %c\n", *option);
					          exit( 1 );
										break;
				}
			}
		}
		else {
			window_title = argv[argc];
		}
	}
}


/* get selected row and exec the editor (or whatever is requested)
 */
static void go_selected( GtkCList *list ) {
	int     i = 0;
	char    *arg;
	rowData *rdata;
	GList *dlist;
	gint   the_row = -1;

	/* fetch the currently selected list item */
	dlist = list->selection;
	if (dlist) {
		the_row = (gint)dlist->data;
	}
	else {
		/* no selection - use focus row instead */
    the_row = list->focus_row;
	}

	if ( the_row < 0 ) {
		return;
	}

	rdata = (rowData *)gtk_clist_get_row_data(list, the_row);
	if ( !rdata ) {
		return;
	}

	/* shutdown the gui */
	gtk_main_quit();

	/* act depending on mode */
	switch ( mode ) {

		case 'd':
			g_print( "File name : %s\n", rdata->fileName );
			g_print( "Regex/Line: %s\n", rdata->searchKey );
			break;				

		case 'l':
			g_print( "%s", rdata->searchKey );
			break;				

		default:		
			/* build the execution arguments (replace proxies) */
			while ( (arg = exec_argv[i]) ) {
				if ( strcmp(arg, "%key") == 0 ) {
					exec_argv[i] = rdata->searchKey;
				}
				else if ( strcmp(arg, "%file") == 0 ) {
					exec_argv[i] = rdata->fileName;
				}
				i++;
			}
			/* execute and die */
			execvp( exec_argv[0], exec_argv );
			break;
	}
}


/* CList double click handler.
 */
static void row_selected( GtkWidget      *clist,
       				            gint            row,
              				    gint            column,
								     			GdkEventButton *event,
                     			gpointer        data )
{
 	if (event && (event->type==GDK_2BUTTON_PRESS) ) {
		go_selected( GTK_CLIST(clist) );
	}
}


/* Key signal handler. Escape will always abort gTags. Return on
 * list items will initiate tag lookup.
 */
static gint key_pressed( GtkWidget *widget, GdkEventKey *event, 
												            gpointer func_data) {

	if ( event->keyval == GDK_Return ) {
		go_selected( GTK_CLIST(func_data) );
	}
	else if ( event->keyval == GDK_Escape ) {
		gtk_main_quit();
	}
	return FALSE;
}

/* "Go" button clicked handler.
 */
static void button_clicked( GtkWidget *gtklist, gpointer func_data ) {
	go_selected( GTK_CLIST(func_data) );
}


/* Returns a format used to display the tag or null, if no format is available.
 */
static char *get_format( char* filename, char hint ) {
	char *ext;
	tagFormat *formats;
	langMap   *langmap;
	
	/* determine file extension */
	if ( ! (ext = strrchr( filename, '.' )) ) {
		return NULL;
	}
	ext++;

	/* find language formats */
	langmap = language_map;
	while ( 1 ) {
		if ( langmap->ext == NULL ) {
			return NULL;
		}
		if ( strcmp(langmap->ext,ext) == 0 ) {
			formats = langmap->formats;
			break;
		}
		langmap++;
	}
	
	/* select format for current hint */
	while ( formats->hint != '\0' ) {
		if ( formats->hint == hint ) {
			return formats->format;
		}
		formats++;
	}
	return NULL;
}

/* makes a copy of a string
 */
static char* copy_of( char *data ) {
	gchar *data_copy;
	data_copy = (gchar *)g_malloc( sizeof(gchar)*(strlen(data) + 1) );
	strcpy( data_copy, data );
	return data_copy;
}

/* allocates a new rowData object
 */
static rowData* newRowData( char *fileName, char *searchKey ) {
	rowData *data = (rowData *)g_malloc( sizeof(rowData) );
	data->fileName  = copy_of( fileName );
	data->searchKey = copy_of( searchKey );
	return data;	
}


/* This reads the tags index from stdin (output of ctags) and appends a row
 * to the passed CList widget for each tag.
 */
static void read_ctags( GtkCList *list ) {

	char  line_buffer[BUFFER_SIZE];
	char *sep;
	char *tag;
	char *filename;
	char *searchkey;
	char *format;
	int   has_format;
	
	char  format_buffer[BUFFER_SIZE];
	char *row_text[1] = { format_buffer };
	int   cur_row = 0;

	while ( fgets(line_buffer,BUFFER_SIZE,stdin) ) {

		/* 1. separator (required) */
		if ( ! (sep = strchr( line_buffer, '\t' )) ) {
			continue;
		}
		*(sep++) = '\0';

		/* 1. field: display tag */
		tag = line_buffer;

		/* 2. field: file name */
		filename = sep;

		/* 2. separator (required) */
		if ( ! (sep = strchr( sep, '\t' )) ) {
			continue;
		}
		*(sep++) = '\0';

		/* 3. field: EX command (line number or regex) */
		searchkey = sep;
		
		/* 3. separator - EX comment (optional) */
		if (sep = strstr( sep, ";\"\t" )) {
			*sep = '\0';
			/* NOTE: we assume that the first character of the extension tags
			 * indicates the "kind" extension without the "kind" key. sep now
			 * points to this character.
			 */
			sep += 3;
		}

		/* DEBUGGING 
		printf( "%s - %s - %s - %c\n", tag, filename, searchkey, *sep );
		*/
		
		/* Formatting */
		format = NULL;
		if (sep) {
			format = get_format( filename, *sep );
		}
		if ( format ) {
			sprintf( format_buffer, format, tag );
		}
		else {
			strcpy( format_buffer, tag );
		}

		/* now add a row to the CList widget */
		gtk_clist_append( list, row_text );
		
		gtk_clist_set_row_data( list, cur_row++, 
														newRowData(filename,searchkey) );
	}
}


/* Life starts here.
 */
gint main ( int argc, gchar *argv[] ) {

    GtkWidget       *window;
    GtkWidget       *scrwin;
    GtkWidget       *vbox;
    GtkWidget       *hbox;
    GtkWidget       *list;
    GtkWidget       *go_button;
    GtkWidget       *close_button;
    
    /* initialize gtk+ (and subsequently gdk) */
    gtk_init(&argc, &argv);
    
		/* parse args */
		parse_args( argc, argv );
		
    /* create the main window. */
    window=gtk_window_new( GTK_WINDOW_TOPLEVEL );
			if (window_title) {
				gtk_window_set_title( GTK_WINDOW(window), window_title );
			}
			else {
				gtk_window_set_title( GTK_WINDOW(window), "gTags" );
			}

    /* a box to arrange the widgets vertically */
    vbox = gtk_vbox_new(FALSE, 5);
    	gtk_container_border_width( GTK_CONTAINER(vbox), 5 );
    	gtk_container_add( GTK_CONTAINER(window), vbox );
    	gtk_widget_show( vbox );
    
    /* scolled window for the listbox */
    scrwin = gtk_scrolled_window_new( NULL, NULL );
			gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(scrwin),
                            	GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
    	gtk_widget_set_usize( scrwin, 250, 350 );
   		gtk_box_pack_start( GTK_BOX(vbox), scrwin, TRUE, TRUE, 0 );		
	   	/* gtk_container_add( GTK_CONTAINER(window), scrwin ); */
    	gtk_widget_show( scrwin );
    
    /* create the CList widget */
    list=gtk_clist_new(1);
			gtk_clist_set_column_width( GTK_CLIST(list), 0, 200 );
																			 		
		/* fill the list from the ctags output*/
		read_ctags( GTK_CLIST(list) );
	
  	/* list selection handler */
	  gtk_signal_connect(GTK_OBJECT(list), "select_row", GTK_SIGNAL_FUNC(row_selected),	NULL);
		gtk_container_add(GTK_CONTAINER(scrwin), list);
  	gtk_widget_show(list);
			
    /* Create hbox for buttons */
    hbox = gtk_hbox_new(TRUE, 5);
   		gtk_box_pack_start( GTK_BOX(vbox), hbox, FALSE, TRUE, 0 );		
    	gtk_widget_show( hbox );

    /* Go button */
    go_button = gtk_button_new_with_label("Go");
    	gtk_container_add( GTK_CONTAINER(hbox), go_button );
    	gtk_widget_show( go_button );
    	gtk_signal_connect( GTK_OBJECT(go_button), "clicked",
		                    	GTK_SIGNAL_FUNC( button_clicked ), list );

    /* Close button */
    close_button = gtk_button_new_with_label("Close");
    	gtk_container_add( GTK_CONTAINER(hbox), close_button );
    	gtk_widget_show( close_button );
    	gtk_signal_connect_object( GTK_OBJECT(close_button), "clicked",
			      									 	 GTK_SIGNAL_FUNC(gtk_widget_destroy),
			      									   GTK_OBJECT(window) );

   	/* close window signal */
   	gtk_signal_connect( GTK_OBJECT(window), "destroy",
		  	                	GTK_SIGNAL_FUNC( gtk_main_quit ), NULL );
		/* escape key handler */
		gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
               			  	GTK_SIGNAL_FUNC( key_pressed ), list);

    /* finaly we want to see the window, dont we? ;) */
    gtk_widget_show(window);

		/* put the list in control */
		gtk_widget_grab_focus( list );

    /* fire up the main event loop of gtk */
    gtk_main();

    return 0;
}





